#!/usr/bin/python
# -*- coding: utf-8 -*-
# author:   Jan Hybs
from loggers import printf
from scripts.core.base import Printer, IO
from scripts.core.pypy import PyPy
from utils.dotdict import Map
from utils.strings import format_n_lines


class RuntestParser(object):
    """
    Class RuntestParser is module for parsing json file result for command runtest
    keyword is test-case
    """

    def __init__(self, obj, batch):
        self.n_lines = 0 if batch else 15
        self.batch = batch
        self.pypy = self.create_from_fields(
            obj.get('execution', {}),
            returncode=int,
            log=str,
            name=str,
            case=lambda x: self.create_from_fields(x, cpu=int, test_case=str)
        )
        self.clean = self.create_from_fields(
            obj.get('clean', {}),
            returncode=int,
            name=str,
            dir=str,
            error=str
        )
        self.comp = self.create_from_fields(
            obj.get('compare', {}),
            returncode=int,
            name=str,
            tests=lambda x: [Map(y) for y in x]
        )
        self.returncode = int(obj.get('returncode'))

    def get_result(self):
        if self.clean.returncode != 0:
            printf.out(
                "{} Could not clean directory '{c[dir]}': {c[error]}",
                self.get_status_line(self.clean),
                c=self.clean)
            return

        if self.pypy.returncode != 0:
            printf.out("{} Run error, case: {p[name]}", self.get_status_line(self.pypy), p=self.pypy)
            self.print_log_file(self.pypy.log, self.n_lines)
            return
        elif self.batch:
            self.print_log_file(self.pypy.log, self.n_lines)

        if self.comp.returncode not in (0, None):
            printf.out("{} Compare error, case: {p[name]}, Details: ", self.get_status_line(self.comp), p=self.pypy)
            self.print_log_file(self.pypy.log, self.n_lines)
            with printf:
                with printf:
                    for c in self.comp.tests:
                        rc = c.returncode
                        if rc == 0:
                            printf.out('[{:^6}]: {}', 'OK', c.name)
                        else:
                            printf.out('[{:^6}]: {}', 'FAILED', c.name)
            return
        elif self.batch:
            self.print_log_file(self.comp.log, self.n_lines)

    @classmethod
    def create_from_fields(cls, json, **fields):
        obj = Map()
        for k, v in list(fields.items()):
            value = json.get(k, None)
            obj[k] = v(value) if value is not None else None
        return obj

    @classmethod
    def get_status_line(cls, o, map=False):
        if not map:
            return '[{:^6}]:{o.returncode:3} |'.format('ERROR', o=o)
        return '[{:^6}]:{o.returncode:3} |'.format(
            PyPy.returncode_map.get(str(o['returncode']), 'ERROR'), o=o)

    @classmethod
    def print_log_file(cls, f, n_lines):
        log_file = IO.read(f)
        if log_file:
            if n_lines == 0:
                printf.out('Full log from file {}:', f)
            else:
                printf.out('Last {} lines from file {}:', abs(n_lines), f)

            printf.stream(format_n_lines(log_file.rstrip(), -n_lines))


class ExecParser(object):
    """
    Class ExecParser is module for parsing json file result for executing command
    keyword is exec
    """

    def __init__(self, obj, batch):
        self.returncode = obj.get('returncode', None)
        self.name = obj.get('name', None)
        self.log = obj.get('log', None)
        self.n_lines = 0 if batch else 15
        self.batch = batch

    def get_result(self):
        if self.returncode != 0:
            if self.returncode != 0:
                printf.out("{} Run error, case: {p.name}", RuntestParser.get_status_line(self), p=self)
                RuntestParser.print_log_file(self.log, self.n_lines)
                return
        elif self.batch:
            RuntestParser.print_log_file(self.log, self.n_lines)


class JsonParser(object):
    """
    Class JsonParser parses json result
    :type tests : list[RuntestParser] or list[ExecParser]
    """
    parse_map = {
        'test-case': RuntestParser,
        'exec': ExecParser
    }

    def __init__(self, obj, batch):
        self.batch = batch
        self.returncode = int(obj.get('returncode'))
        self.error = obj.get('error')
        self.tests = list()

        for t in obj.get('tests', []):
            self.tests.append(self.parse_map.get(t['type'])(t, batch))
